#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Elasticsearch service.
"""

import json
import http.client
import logging
import os
import socket
import urllib.parse
import urllib.request

from forklift.base import open_root_owned
from .base import (cache_directory,
                   container_name_for,
                   ProviderNotAvailable,
                   pipe_split,
                   replace_part,
                   register,
                   URLNameDescriptor,
                   URLService,
                   transient_provider)

LOGGER = logging.getLogger(__name__)


try:
    # pylint:disable=undefined-variable,invalid-name
    CONNECTION_ISSUES_ERROR = ConnectionError
except NameError:
    # pylint:disable=invalid-name
    CONNECTION_ISSUES_ERROR = urllib.error.URLError


@register('elasticsearch')
class Elasticsearch(URLService):
    """
    Elasticsearch service for the application.
    """

    allow_override = URLService.allow_override + ('index_name',)

    providers = ('localhost', 'container')

    CONTAINER_IMAGE = 'dockerfile/elasticsearch'

    DEFAULT_PORT = 9200

    index_name = URLNameDescriptor()

    TEMPORARY_AVAILABILITY_ERRORS = (CONNECTION_ISSUES_ERROR,
                                     http.client.HTTPException,
                                     ProviderNotAvailable,
                                     socket.error,
                                     ValueError)
    PERMANENT_AVAILABILITY_ERRORS = (urllib.request.URLError,)

    def __init__(self, index_name, urls):
        super().__init__(tuple(
            urllib.parse.urljoin(url, index_name)
            for url in pipe_split(urls)
        ))

    def environment(self):
        """
        The environment to access Elasticsearch.
        """

        hosts = '|'.join(
            replace_part(url, path='').geturl()
            for url in self.urls
        )
        index_name = self.urls[0].path[1:]

        return {
            'ELASTICSEARCH_INDEX_NAME': index_name,
            'ELASTICSEARCH_URLS': hosts,
        }

    def check_available(self):
        """
        Check whether Elasticsearch is available at a given URL.
        """

        if not self.urls:
            return False

        for url in self.urls:
            url = replace_part(url, path='')
            es_response = urllib.request.urlopen(url.geturl())
            es_status = json.loads(es_response.read().decode())
            if es_status['status'] != 200:
                raise ProviderNotAvailable(
                    ("Provider '{}' is not yet available: HTTP response "
                     "{}\n{}").format(self.__class__.__name__,
                                      es_status['status'],
                                      es_status['error'])
                )

        return True

    @classmethod
    def localhost(cls, application_id):
        """
        The Elasticsearch environment on the local machine.
        """
        return cls(index_name=application_id,
                   urls=('http://localhost:9200',))

    @classmethod
    @transient_provider
    def container(cls, application_id):
        """
        Elasticsearch provided by a container.
        """

        image_name = cls.CONTAINER_IMAGE
        container_name = container_name_for(image_name, application_id)
        cache_dir = cache_directory(container_name)

        if not os.path.exists(cache_dir):
            LOGGER.debug("Creating cache directory '%s'", cache_dir)
            os.makedirs(cache_dir)

        config_path = os.path.join(cache_dir, 'elasticsearch.yml')
        LOGGER.debug("Writing ElasticSearch config to '%s'", config_path)
        with open_root_owned(config_path, 'w') as config:
            print(
                """
                path:
                    data: /data/data
                    logs: /data/log
                """,
                file=config,
            )

        container = cls.ensure_container(application_id, data_dir='/data')

        instance = cls(
            index_name=application_id,
            urls=('http://{host}:{port}'.format(**container.__dict__),),
        )
        instance.wait_until_available()
        # pylint:disable=attribute-defined-outside-init
        instance.container_info = container
        return instance



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PostgreSQL database service.
"""

import logging
import os
import re
import subprocess

from forklift.base import DEVNULL
from .base import (ProviderNotAvailable,
                   URLNameDescriptor,
                   URLService,
                   register,
                   transient_provider)


LOGGER = logging.getLogger(__name__)


@register('postgres')
class PostgreSQL(URLService):
    """
    PostgreSQL service provided by the host machine.
    """

    CHECK_COMMAND = 'select version()'
    CONTAINER_IMAGE = 'paintedfox/postgresql'
    DATABASE_NAME = 'DEFAULT'
    DEFAULT_PORT = 5432
    URL_SCHEME = 'postgres'

    TEMPORARY_AVAILABILITY_ERRORS = (ProviderNotAvailable,)
    PERMANENT_AVAILABILITY_ERRORS = (ProviderNotAvailable,
                                     subprocess.CalledProcessError,
                                     OSError)

    allow_override = URLService.allow_override + ('name',)

    name = URLNameDescriptor()

    # pylint:disable=too-many-arguments
    def __init__(self,
                 name,
                 host='localhost',
                 port=DEFAULT_PORT,
                 user=None,
                 password=None):
        super().__init__((
            '{scheme}://{user}{password}@{host}:{port}/{name}'.format(
                scheme=self.URL_SCHEME,
                user=user,
                password=':' + password if password else '',
                host=host,
                port=port,
                name=name,
            ),
        ))

    def environment(self):
        """
        The environment needed for the application to connect to PostgreSQL.
        """

        env_name = 'DB_{0}_URL'.format(self.DATABASE_NAME)
        return {env_name: self.urls[0].geturl()}

    def check_available(self):
        """
        Check whether PostgreSQL is installed on the host and accessible. Will
        raise ProviderNotAvailable or subprocess.CalledProcessError when
        unavailable
        """

        stderr = ""
        subprocess_kwargs = {
            'stdin': DEVNULL,
            'stdout': DEVNULL,
            'stderr': subprocess.PIPE,
        }

        def get_proc_stderr(proc):
            """
            Safely read data from stderr into a string and return it
            """
            proc_stderr = ""
            for stderr_data in proc.communicate():
                proc_stderr += str(stderr_data)
            proc.wait()
            return proc_stderr

        psql_check_command = ['psql', '--version']
        proc = subprocess.Popen(psql_check_command, **subprocess_kwargs)
        stderr = get_proc_stderr(proc)

        if proc.returncode != 0:
            raise subprocess.CalledProcessError(
                returncode=proc.returncode,
                cmd=' '.join(psql_check_command),
                output=stderr
            )

        if self.password:
            os.environ['PGPASSWORD'] = self.password

        proc = subprocess.Popen([
            'psql',
            '-h', self.host,
            '-p', str(self.port),
            '-U', self.user,
            '-w',
            self.name,
            '-c', self.CHECK_COMMAND,
        ], **subprocess_kwargs)
        stderr += get_proc_stderr(proc)

        if proc.returncode != 0:
            raise ProviderNotAvailable(
                ("Provider '{}' is not yet available: psql exited with status "
                 "{}\n{}").format(self.__class__.__name__,
                                  proc.returncode,
                                  stderr)
            )

        return True

    @classmethod
    def localhost(cls, application_id):
        """
        The PostgreSQL environment on the local machine.
        """
        return cls(
            host='localhost',
            name=application_id,
            user=application_id,
        )

    @classmethod
    @transient_provider
    def container(cls, application_id):
        """
        PostgreSQL provided by a container.
        """

        db_name = re.sub(r'[^a-zA-Z0-9_]', '_', application_id)

        container = cls.ensure_container(
            application_id,
            # FIXME: this is broken at the moment in the paintedfox container
            # data_dir='/data',
            environment={
                'DB': db_name,
                'PASS': 'forklift',
            }
        )

        instance = cls(
            host=container.host,
            port=container.port,
            name=db_name,
            user='super',
            password='forklift',
        )
        instance.wait_until_available()
        # pylint:disable=attribute-defined-outside-init
        instance.container_info = container
        return instance

    providers = ('localhost', 'container')


@register('postgis')
class PostGIS(PostgreSQL):
    """
    PostgreSQL database with PostGIS support.
    """

    CHECK_COMMAND = """CREATE EXTENSION IF NOT EXISTS postgis;
                       SELECT PostGIS_full_version()"""
    CONTAINER_IMAGE = 'thatpanda/postgis'
    URL_SCHEME = 'postgis'

    providers = ('localhost', 'container')



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Memcache service.
"""

from .base import (
    ProviderNotAvailable,
    URLHostInfoDescriptor,
    URLNameDescriptor,
    URLService,
    port_open,
    register,
    split_host_port,
)


@register('memcache')
class Memcache(URLService):
    """
    Memcache service for the application.
    """

    DEFAULT_PORT = 11211

    CONTAINER_IMAGE = 'fedora/memcached'

    allow_override = URLService.allow_override + ('key_prefix',)
    allow_override_list = URLService.allow_override_list + ('hosts',)
    key_prefix = URLNameDescriptor()
    hosts = URLHostInfoDescriptor(default_port=DEFAULT_PORT, joiner=tuple)

    providers = ('localhost', 'container')

    def __init__(self,
                 key_prefix='',
                 hosts=None):

        super().__init__(
            'memcache://{host}:{port}/{key_prefix}'.format(
                host=host,
                port=port,
                key_prefix=key_prefix,
            )
            for host, port in (
                split_host_port(h, self.DEFAULT_PORT)
                for h in hosts
            )
        )

    def environment(self):
        """
        The environment to access Memcache
        """

        return {
            'MEMCACHE_HOSTS': '|'.join(self.hosts),
            'MEMCACHE_PREFIX': self.key_prefix,
        }

    def check_available(self):
        """
        Check whether memcache is available

        Do this by connecting to the socket. At least one host must be up
        """

        for host in self.hosts:
            if port_open(*split_host_port(host, self.DEFAULT_PORT)):
                return True

        raise ProviderNotAvailable("Memcached not available: none of the hosts"
                                   " are up")

    @classmethod
    def localhost(cls, application_id):
        """
        The default memcached provider
        """

        return cls(key_prefix=application_id,
                   hosts=['localhost:{0}'.format(cls.DEFAULT_PORT)])

    @classmethod
    def from_container(cls, application_id, container):
        """
        Memcached provided by a container.
        """

        return cls(
            key_prefix=application_id,
            hosts=['{host}:{port}'.format(**container.__dict__)],
        )



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Syslog service.
"""

import socket

from forklift.base import free_port
from .base import (
    register,
    transient_provider,
    try_port,
    URLDescriptor,
    URLService,
)


@register('syslog')
class Syslog(URLService):
    """
    Logging facility for the application.
    """

    DEFAULT_PORT = 514

    TEMPORARY_AVAILABILITY_ERRORS = (socket.error,)

    allow_override = ('host', 'port', 'proto')

    proto = URLDescriptor('scheme')

    def __init__(self, host=None, port=DEFAULT_PORT, proto='udp'):
        super().__init__('{proto}://{host}:{port}'.format(
            host=host or '',
            port=port,
            proto=proto,
        ))

    def environment(self):
        """
        The environment to provide logging.
        """

        return {
            'SYSLOG_SERVER': self.host,
            'SYSLOG_PORT': str(self.port),
            'SYSLOG_PROTO': self.proto,
        }

    def check_available(self):
        """
        Check whether syslog is available.

        If the protocol is UDP, assume it is available if any of the other
        parameters are set.
        """

        if not self.host:
            return False

        if self.proto == 'udp':
            return True
        else:
            return try_port(self.host, self.port)

    # pylint:disable=unused-argument
    @classmethod
    def manual(cls, application_id):
        """
        Manually-configured syslog. Will not be available unless parameters
        are overridden in configuration.
        """

        return cls()

    # pylint:disable=unused-argument
    @classmethod
    @transient_provider
    def stdout(cls, application_id):
        """
        Logger printing all the messages to the standard output of Forklift.
        """

        # Adapted from https://gist.github.com/marcelom/4218010
        from forklift.services.satellite import start_satellite
        import socketserver

        class SyslogHandler(socketserver.BaseRequestHandler):
            """
            Handler outputting logging messages received to stdout.
            """
            def handle(self):
                data = self.request[0].strip().decode()
                print(data)

        port = free_port()

        def run_server():
            """
            Run the syslog server.
            """

            server = socketserver.UDPServer(('0.0.0.0', port), SyslogHandler)
            server.serve_forever()

        start_satellite(target=run_server)

        return cls('localhost', port)

    providers = ('manual', 'stdout')



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Satellite processes started by Forklift itself to provide services.
"""

import os
from threading import Thread

from forklift.base import wait_for_parent


def start_satellite(target, args=(), kwargs=None, stop=None):
    """
    Start a process configured to run the target but kill it after the parent
    exits.
    """

    if kwargs is None:
        kwargs = {}

    child_pid = os.fork()
    if not child_pid:
        # Make sure signals sent by the shell aren't propagated to the
        # satellite
        os.setpgrp()
        _satellite(target, args, kwargs, stop)


def _satellite(target, args, kwargs, stop):
    """
    Run the target, killing it after the parent exits.
    """

    # Run target daemonized.
    payload = Thread(
        target=target,
        args=args,
        kwargs=kwargs,
    )
    payload.daemon = True
    payload.start()

    wait_for_parent()
    exit_status = stop() if stop is not None else None
    if exit_status is None:
        exit_status = 0

    # This is in a child process, so exit without additional cleanup
    os._exit(exit_status)  # pylint:disable=protected-access



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common declarations.
"""

import os
import socket
import subprocess
import tempfile
import time

import psutil

from contextlib import contextmanager


def free_port():
    """
    Find a free TCP port.
    """

    with socket.socket() as sock:
        sock.bind(('', 0))
        return sock.getsockname()[1]


def wait_for(func, expected_exceptions=(), retries=60):
    """
    Wait for a function to return a truthy value, possibly ignoring some
    exceptions if they are raised until the very last retry

    Parameters:
        func - the function to continually call until truthy
        expected_exceptions - list of exceptions to ignore, unless the final
            retry is reached (then any exceptions are reraised)
        retries - number of times to retry before giving up

    Return value:
        The return value of func the last time it was run
    """

    retries = int(retries)
    for retry in range(1, retries + 1):
        try:
            return_value = func()
            if return_value:
                break

        except expected_exceptions:
            if retry == retries:
                raise
            else:
                pass

        time.sleep(1)

    return return_value


def wait_for_parent():
    """
    Use wait_for_pid to wait for your parent process
    """
    wait_for_pid(os.getppid())


def wait_for_pid(pid):
    """
    Wait for a given PID in the best way possible. If PID is a child, we use
    os.waitpid. Otherwise, we fall back to a polling approach.
    """
    try:
        # Try to wait for a child
        os.waitpid(pid, 0)
    except OSError:
        # Fallback to polling process status
        try:
            proc = psutil.Process(pid)
            while proc.status() not in ('zombie', 'dead'):
                time.sleep(1)
        except psutil.NoSuchProcess:
            pass


@contextmanager
def open_root_owned(source, *args, **kwargs):
    """
    Copy a file as root, open it for writing, then copy it back as root again
    when done
    """
    with tempfile.NamedTemporaryFile(*args, **kwargs) as dest_fh:
        if os.path.isfile(source):
            subprocess.check_call(['sudo', 'cp', source, dest_fh.name])
        yield dest_fh
        subprocess.check_call(['sudo', 'cp', dest_fh.name, source])


def rm_tree_root_owned(path):
    """
    Do an equivalent of shutil.rmtree, but as root
    """
    subprocess.check_call(['sudo', 'rm', '-rf', path])


class ImproperlyConfigured(Exception):
    """
    The host is not properly configured for running Docker.
    """

    pass

try:
    DEVNULL = subprocess.DEVNULL  # pylint:disable=no-member
except AttributeError:
    DEVNULL = open(os.devnull)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Proxy service.
"""

from .base import URLService, port_open, register, transient_provider


@register('proxy')
class Proxy(URLService):
    """
    Proxy service for the application.
    """

    def __init__(self, host='', port=None):
        super().__init__('http://{host}{port}'.format(
            host=host or '',
            port=':' + str(port) if port else '',
        ))

    def available(self):
        """
        Check whether the proxy is available.
        """

        if self.host:
            return port_open(self.host, self.port)
        else:
            return True

    def environment(self):
        """
        The environment to access the proxy.
        """

        if self.host:
            return {
                'HTTP_PROXY': self.url_string()
            }
        else:
            return {}

    # pylint:disable=unused-argument
    @classmethod
    @transient_provider
    def manual(cls, application_id):
        """
        Manually-configured proxy. Will not be available unless parameters
        are overridden in configuration.
        """

        return cls()

    providers = ('manual',)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Drivers that can execute applications.
"""


import fcntl
import os
import pwd
import re
import socket
import struct
import subprocess
import sys

# pylint:disable=no-name-in-module,import-error
from distutils.spawn import find_executable
# pylint:enable=no-name-in-module,import-error

import docker

from forklift.base import (
    DEVNULL,
    free_port,
    ImproperlyConfigured,
    wait_for,
    wait_for_parent,
)
from forklift.registry import Registry


register = Registry()  # pylint:disable=invalid-name


class Driver(object):
    """
    A method of executing the application with supplied services.
    """

    def __init__(self, target, services, environment, conf):
        """
        Initialise the driver with the specified target and services.
        """

        self.target = target
        self.services = services
        self.added_environment = environment
        self.conf = conf

    # pylint:disable=unused-argument
    @staticmethod
    def valid_target(target):
        """
        Guess whether a target is valid for the given driver.

        Override to have driver chosen automatically based on target
        given.
        """

        return False

    def run(self, *command):
        """
        Run the command on the target.
        """

        raise NotImplementedError("Please override run().")

    def _run(self, command):
        """
        Run the command AS THE CURRENT PROCESS (you will be replaced).

        We also fork and add the child to its own process group. This means
        that the child will persist after the parent completes and will not
        receive signals (SIGTERM, SIGKILL, SIGSTOP, etc) from the parent
        pgroup, and will also not grab standard input from the parent process.

        When the parent process finishes, we return (in the child process) so
        that the main code path can complete
        """
        child_pid = os.fork()
        if child_pid:
            os.execvp(command[0], command)
        else:
            os.setpgrp()
            wait_for_parent()

    def base_environment(self):
        """
        The service-independent environment to supply to the application.
        """

        return {
            'ENVIRONMENT': 'dev_local',
            'DEVNAME': pwd.getpwuid(os.getuid())[0],
            # TODO: TZ
            'SITE_PROTOCOL': 'http',
            'SITE_DOMAIN': 'localhost:{0}'.format(self.serve_port()),
        }

    def environment(self):
        """
        The environment to supply to the application.
        """

        env = self.base_environment()

        for service in self.services:
            env.update(service.environment())

        env.update(self.added_environment)

        return env

    @staticmethod
    def _free_port():
        """
        Find a free port to serve on.

        Overridden in tests for reproducibility.
        """

        return free_port()

    def serve_port(self):
        """
        The port for application to serve in.

        If not explicitly given in the configuration, pick a free one.
        """

        if self.conf.serve_port:
            return self.conf.serve_port

        # pylint:disable=access-member-before-definition
        # pylint:disable=attribute-defined-outside-init
        if hasattr(self, '_serve_port'):
            return self._serve_port

        self._serve_port = self._free_port()
        return self._serve_port

    @classmethod
    def add_arguments(cls, add_argument):
        """
        Add driver configuration arguments to the parser.
        """

        add_argument('--serve_port', type=int, default=None,
                     help="The port to expose the application on")


def ip_address(ifname):
    """
    Get the IP address associated with an interface.
    """

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as tmp_socket:
        return socket.inet_ntoa(fcntl.ioctl(
            tmp_socket.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', ifname[:15].encode())
        )[20:24])


@register('docker')
class Docker(Driver):
    """
    Execute the application packaged as a Docker container.
    """

    @classmethod
    def add_arguments(cls, add_argument):
        """
        Add Docker-specific options.
        """

        super().add_arguments(add_argument)

        add_argument('--rm', default=False, action='store_true',
                     help="Remove the container after the command exit")
        add_argument('--privileged', default=False, action='store_true',
                     help="Run the container in privileged mode")
        add_argument('--interactive', default=False, action='store_true',
                     help="Run the command in interactive mode")
        add_argument('--mount-root',
                     help="The directory to bind the root directory of the " +
                     "container to")
        add_argument('--storage',
                     help="The directory to mount under /storage in the " +
                     "container")
        add_argument('--user', default='app',
                     help="The user to set up for SSH in the container")
        add_argument('--host_uid', default=os.getuid(),
                     help="The UID for the user inside the container")
        add_argument('--identity',
                     help="The public key to authorise logging in as")

    def run(self, *command):
        """
        Run the command in Docker container.
        """

        # Check resolv.conf for local nameservers
        with open('/etc/resolv.conf') as rcfile:
            nameservers = [l for l in rcfile.read().splitlines()
                           if 'nameserver' in l and not l.startswith('#')]
            if any('127.0.0.1' in l or '127.0.1.1' in l
                   for l in nameservers):
                print(
                    "/etc/resolv.conf on the host specifies localhost "
                    "as the name server. This will make Docker use Google "
                    "Public DNS inside the container, and accessing Intranet "
                    "resources will fail.\n"
                    "Please fix /etc/resolv.conf on the host before "
                    "continuing.",
                    file=sys.stderr,
                )

        if list(command) == ['sshd']:
            return self.run_sshd()
        else:
            command = self.docker_command(*command)
            return self._run(command)

    def environment(self):
        """
        Change every service's host attribute from localhost.
        """

        for service in self.services:
            if 'host' in service.allow_override:
                if service.host == 'localhost':
                    service.host = ip_address('docker0')

        return super().environment()

    def docker_command(self, *command, use_sshd=False):
        """
        The Docker command to start a container.
        """

        docker_command = [
            'docker', 'run',
            '-p', '{0}:8000'.format(self.serve_port()),
        ]

        if self.conf.rm:
            docker_command += ['--rm']

        if use_sshd:
            docker_command += [
                '-d',
                '-p', '22',
                '--entrypoint=/bin/bash',
                '-u=root',
            ]
        else:
            for key, value in self.environment().items():
                docker_command += ['-e', '{0}={1}'.format(key, value)]

        if self.conf.privileged:
            docker_command += [
                '--privileged',
            ]

        if self.conf.interactive:
            docker_command += [
                '-i', '-t',
            ]

        if self.conf.storage:
            storage = self.conf.storage
            subprocess.check_call(['mkdir', '-p', storage])
            docker_command += [
                '-v', '{0}:/storage'.format(storage),
            ]
        docker_command += [self.target]
        docker_command += command

        return docker_command

    def mount_root(self, container):
        """
        Mount the container's root directory on the host.
        """

        # If requested, mount the working directory
        if self.conf.mount_root:
            mount_root = self.conf.mount_root

            subprocess.call(['sudo', 'umount', mount_root],
                            stderr=DEVNULL)
            subprocess.check_call(['mkdir', '-p', mount_root])

            with docker.Client() as client:
                driver = client.inspect_container(container)['Driver']
            rootfs_path = \
                '/var/lib/docker/{driver}/mnt/{container}'.format(
                    driver=driver,
                    container=container,
                )

            # AUFS and DeviceMapper use different paths
            if driver == 'devicemapper':
                rootfs_path += '/rootfs'

            subprocess.check_call(['sudo', 'mount', '-o', 'bind',
                                   rootfs_path,
                                   mount_root])
            print("Container filesystem mounted on {mount_root}".format(
                mount_root=mount_root))

    def run_sshd(self):
        """
        Run SSH server in a container.
        """

        # determine the user's SSH key(s)
        identity = None
        if not self.conf.identity:
            # provide the entire set of keys
            # pylint:disable=no-member
            ssh_key = (subprocess
                       .check_output(['ssh-add', '-L'])
                       .decode()
                       .strip())
            if ssh_key == '':
                raise ImproperlyConfigured(
                    "You don't seem to have any SSH keys! "
                    "How do you do any work?")
        else:
            identity = self.conf.identity
            if not os.path.exists(identity):
                identity = os.path.expanduser(
                    '~/.ssh/{}'.format(self.conf.identity))
            with open(identity + '.pub') as id_file:
                ssh_key = id_file.read().strip()

        commands = [
            'DEBIAN_FRONTEND=noninteractive apt-get -qq install dropbear sudo',
            '(kill $(pidof dropbear) || true)',
        ] + [
            # TODO: this only passes the environment to shells.
            # Commands run directly (ssh ... command) get no environment.
            'echo \'export {0}={1}\' >> /etc/profile'.format(*env)
            for env in self.environment().items()
        ] + [
            '(useradd -m {user} || true)',
            'mkdir -p ~{user}/.ssh',
            'echo \'{ssh_key}\' >> ~{user}/.ssh/authorized_keys',
            'chown -R {user} ~{user}/.ssh',
            'chmod -R go-rwx ~{user}/.ssh',
            'chsh -s /bin/bash {user}',
            'usermod -p zzz {user}',
            'chown -R --from={user} {host_uid} ~{user}',
            'usermod -u {host_uid} {user}',
            'chown {user} ~{user}',
            'chmod go-w ~{user}',
            'echo \'{user} ALL=(ALL) NOPASSWD: ALL\' >> /etc/sudoers',
            'echo Starting SSH...',
            '/usr/sbin/dropbear -F',
        ]

        args = {
            'user': self.conf.user,
            'host_uid': self.conf.host_uid,
            'ssh_key': ssh_key,
        }

        command = self.docker_command(
            '-c',
            ' && '.join(cmd.format(**args) for cmd in commands),
            use_sshd=True
        )
        container = subprocess.check_output(command).decode().strip()
        self.mount_root(container)

        ssh_command, ssh_available = self.ssh_command(container, identity)
        if not ssh_available:
            print("Timed out waiting for SSH setup. You can still try "
                  "the command below but it might also indicate a problem "
                  "with SSH setup.")
        print(ssh_command)

        return 0

    def ssh_command(self, container, identity=None):
        """
        Wait for SSH service to start and print the command to SSH to
        the container.

        Returns a tuple of (command, available), where command is the command
        to run and available is an indication of whether the self-test
        succeeded.
        """

        with docker.Client() as docker_client:
            ssh_port = docker_client.port(container, 22)[0]['HostPort']

        ssh_command = [
            'ssh',
            '{0}@localhost'.format(self.conf.user),
            '-p',
            ssh_port,
            '-A',
        ]

        if identity:
            ssh_command += ('-i', identity)

        available = False
        try:
            available = wait_for(
                lambda: subprocess.check_call(
                    ssh_command + [
                        '-o', 'StrictHostKeyChecking=no',
                        '-o', 'PasswordAuthentication=no',
                        '-o', 'NoHostAuthenticationForLocalhost=yes',
                        '/bin/true',
                    ],
                    stdin=DEVNULL,
                    stdout=DEVNULL,
                    stderr=DEVNULL,
                ) or True,
                expected_exceptions=(subprocess.CalledProcessError,)
            )
        except (subprocess.CalledProcessError, OSError) as ex:
            print(ex)

        return (' '.join(ssh_command), available)


@register('container_recycler')
class ContainerRecycler(Driver):
    """
    Cleans up Docker's mess
    """

    @classmethod
    def add_arguments(cls, add_argument):
        """
        Add recycler-specific options.
        """

        super().add_arguments(add_argument)

        add_argument('--include-running', default=False, action='store_true',
                     help="Remove running containers as well")
        add_argument('--include-tagged', default=False, action='store_true',
                     help="Remove tagged images as well")

    def run(self, *command):
        """
        Recycle old containers and images
        """

        self.recycle_containers(include_running=self.conf.include_running)
        self.recycle_images(include_tagged=self.conf.include_tagged)

    def recycle_containers(self, include_running=False):
        """
        Clean up old stopped (and optionally running) containers
        """
        all_containers = set(
            subprocess.check_output(('docker', 'ps', '-aq'))
            .split()
        )

        running_containers = set(
            subprocess.check_output(('docker', 'ps', '-q'))
            .split()
        )

        if include_running:
            containers = all_containers

        else:
            if running_containers:
                print("You have running containers, pass "
                      "--include-running to remove")

            containers = all_containers - running_containers

        if containers:
            print("Removing old containers...")
            subprocess.check_call(('docker', 'rm', '-f') + tuple(containers))

    def recycle_images(self, include_tagged=False):
        """
        Clean up untagged (and optionally tagged) images
        """
        images = set()
        tagged_images = set()

        output = subprocess.check_output(('docker', 'images'),
                                         universal_newlines=True)
        output = output.strip().split('\n')

        # the first line contains the offsets
        header = output[0]
        remainder = output[1:]

        # calculate the column widths from the header by calculating the
        # offsets of the columns
        columns = [header.index(l)
                   for l in re.split(r'\s\s+', header)] + [len(header)]
        columns = [(a, b) for a, b in zip(columns, columns[1:])]

        for line in remainder:
            repo, tag, image, _, _ = (line[a:b].strip() for a, b in columns)
            images.add(image)

            if repo != '<none>' and tag != '<none>':
                tagged_images.add(image)

        if include_tagged:
            pass

        else:
            if tagged_images:
                print("You have tagged images, pass "
                      "--include-tagged to remove")

            images -= tagged_images

        if images:
            print("Removing old images...")
            subprocess.check_call(('docker', 'rmi') + tuple(images))

    @staticmethod
    def valid_target(target):

        return target == 'recycle'


@register('direct')
class Direct(Driver):
    """
    Execute the application directly.
    """

    def run(self, *command):
        """
        Run the application directly.
        """

        for key, value in self.environment().items():
            os.environ[key] = value

        return self._run([self.target] + list(command))

    @staticmethod
    def valid_target(target):
        """
        Check if the target is directly executable.
        """

        return find_executable(target)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helpers related to maintaining a registry of classes
"""


class Registry(dict):
    """
    A registry class, used for registering services, drivers, etc.

    This is not the registry itself. The registry itself is in
    forklift.services, forklift.drivers, etc.
    """

    def __call__(self, name):
        """
        Use registry as a decorator to register Forklift services
        """

        def inner(cls):
            """
            Decorator
            """
            self[name] = cls

            return cls

        return inner



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for help invocation.
"""

import io
import tempfile

from tests.base import (
    TestCase,
    TestForklift,
    redirect_stream,
)


class HelpTestForklift(TestForklift):
    """
    Mock _readme_stream for tests.
    """

    @staticmethod
    def _readme_stream():
        """
        Dummy stream.
        """

        return io.BytesIO("Help yourself.".encode())


class HelpTestCase(TestCase):
    """
    Test help invocation.
    """

    forklift_class = HelpTestForklift

    def test_help(self):
        """
        Test help invocation.
        """

        with tempfile.NamedTemporaryFile() as tmpfile:
            with redirect_stream(tmpfile.file.fileno()):
                self.run_forklift('help')

            with open(tmpfile.name) as saved_output:
                help_text = saved_output.read()
                help_text = help_text.replace('()', '').strip()
                self.assertEqual("Help yourself.", help_text)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test services API.
"""

import unittest

import forklift.services

from tests.base import docker_image_available


class ServiceTestCase(unittest.TestCase):
    """
    Generic service tests.
    """

    service_class = None

    def test_api_conformance(self):
        """
        Test that the service has the correct API.
        """

        # assert we have at least one provider
        self.assertGreaterEqual(len(self.service_class.providers), 1)

        # assert those providers exist on the class
        for provider in self.service_class.providers:
            assert hasattr(self.service_class, provider)

        # assert can build a provider
        service = getattr(self.service_class,
                          self.service_class.providers[0])('fake')

        # assert we can set the host
        #
        # Only the Docker driver uses the host property, and it is
        # currently optional. However this test is useful because the
        # property is useful. If it turns out there are services for
        # which host is not useful, then this test should be changed :)
        assert hasattr(service, 'host')
        service.host = 'badger'

        assert hasattr(service, 'environment')
        assert hasattr(service, 'available')

        # Test all attributes in allow_override exist
        for attr in service.allow_override:
            value = getattr(service, attr)
            setattr(service, attr, value)

    def test_available(self):
        """
        Test that the service provided by the image is available.
        """

        image = self.service_class.CONTAINER_IMAGE
        if image and not docker_image_available(image):
            raise unittest.SkipTest(
                "Docker image {0} is required.".format(image))

        service = self.service_class.provide('fake', transient=True)
        self.assertTrue(service.available())

        service.cleanup()


def load_tests(loader, tests, pattern):  # pylint:disable=unused-argument
    """
    Generate a test class for each service.
    """

    suite = unittest.TestSuite()
    for cls in forklift.services.register.values():
        test_class = type(ServiceTestCase)(
            cls.__name__ + 'TestCase',
            (ServiceTestCase,),
            {
                'service_class': cls,
            }
        )
        suite.addTests(loader.loadTestsFromTestCase(test_class))
    return suite



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Redis service
"""

import logging
import socket
from telnetlib import Telnet

from .base import (
    register,
    URLNameDescriptor,
    URLService,
)

LOGGER = logging.getLogger(__name__)


@register('redis')
class Redis(URLService):
    """
    A Redis service

    This is a single Redis server.
    """

    allow_override = URLService.allow_override + ('db_index',)
    db_index = URLNameDescriptor()

    providers = ('localhost', 'container')

    CONTAINER_IMAGE = 'dockerfile/redis'

    DEFAULT_PORT = 6379

    TEMPORARY_AVAILABILITY_ERRORS = (socket.error,)

    def __init__(self, host, db_index=0):
        # FIXME: we don't support multiple redis servers yet
        super().__init__('redis://{host}/{db_index}'.format(
            host=host,
            db_index=db_index,
        ))

    def environment(self):
        """
        The environment to access Redis
        """

        return {
            'REDIS_HOSTS': self.host,
            'REDIS_DB_INDEX': str(self.db_index),
        }

    def check_available(self):
        """
        Check whether Redis is available
        """

        # pylint:disable=invalid-name
        nc = Telnet(self.host, self.port)

        try:
            nc.write(b'PING')
            nc.read_until(b'PONG\n',
                          timeout=1)
        finally:
            nc.close()

        return True

    # pylint:disable=unused-argument
    @classmethod
    def localhost(cls, application_id):
        """
        The default Redis provider
        """

        return cls(host='localhost:{port}'.format(port=cls.DEFAULT_PORT))

    @classmethod
    def from_container(cls, application_id, container):
        """
        Redis provided by a container.
        """

        return cls(
            host='{host}:{port}'.format(**container.__dict__),
        )



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for Forklift.
"""

import contextlib
import re
import sys
import tempfile
import yaml

import forklift
from forklift.drivers import ip_address
from tests.base import (
    requires_docker,
    parse_environment,
    redirect_stream,
    DOCKER_BASE_IMAGE,
    SaveOutputMixin,
    TestCase,
)


class UsageTestCase(TestCase):
    """
    Test running forklift with no arguments.
    """

    # Do not override any drivers or services
    forklift_class = forklift.Forklift

    def test_usage(self):
        """
        Test usage message with no arguments.
        """
        with tempfile.NamedTemporaryFile() as tmpfile:
            with redirect_stream(tmpfile.file.fileno(), stream=sys.stderr):
                with self.assertRaises(SystemExit):
                    self.run_forklift()

            with open(tmpfile.name) as saved_stderr:
                usage_test = saved_stderr.read()
                self.assertIn("usage: ", usage_test)


class SmokeTestCase(TestCase):
    """
    Test running basic commands.
    """

    def test_commands(self):
        """
        Test running basic commands.
        """

        self.assertEqual(0, self.run_forklift('true'))
        self.assertNotEqual(0, self.run_forklift('false'))


class CommandsMixin(object):
    """
    Mixin with tests to ensure commands are run correctly.
    """

    def run_command(self, *command):
        """
        Run a command in Forklift.

        Override to pass extra options.
        """

        return self.run_forklift(*command)

    def test_exit_code(self):
        """
        Test command exit codes.
        """

        self.assertEqual(0, self.run_command('/bin/true'))
        self.assertNotEqual(0, self.run_command('/bin/false'))

    def test_output(self):
        """
        Test echoing things.
        """

        self.assertEqual(0, self.run_command('/bin/echo', 'apple', 'orange'))
        self.assertEqual('apple orange\n', SaveOutputMixin.last_output())

        self.assertEqual(
            0,
            self.run_command('--', '/bin/echo', '--apple', '--orange')
        )
        self.assertEqual('--apple --orange\n', SaveOutputMixin.last_output())


class DirectCommandsTestCase(CommandsMixin, TestCase):
    """
    Test running commands directly.
    """

    default_driver = 'direct'


@requires_docker
class DockerCommandsTestCase(CommandsMixin, TestCase):
    """
    Test running commands via Docker.
    """

    default_driver = 'docker'

    def run_command(self, *command):
        """
        Run a command via Docker.
        """

        return self.run_forklift(
            '--rm',
            DOCKER_BASE_IMAGE,
            *command
        )


class CaptureEnvironmentMixin(object):
    """
    Mixin with tests to ensure environment is passed to commands correctly.
    """

    def capture_env(self, *args, prepend_args=None):
        """
        Run Forklift to capture the environment.
        """

        prepend_args = prepend_args or []

        if any(arg.startswith('--') for arg in prepend_args):
            prepend_args.insert(0, '--')

        forklift_args = \
            list(args) + \
            list(prepend_args) + \
            ['/usr/bin/env', '-0']

        self.assertEqual(0, self.run_forklift(*forklift_args))

        return parse_environment(SaveOutputMixin.last_output())

    @contextlib.contextmanager
    def configuration_file(self, configuration):
        """
        Run a command with configuration written to the configuration file.
        """

        with tempfile.NamedTemporaryFile() as conffile:
            self.forklift_class.configuration_file_list.append(conffile.name)
            if isinstance(configuration, str):
                conffile.write(configuration.encode())
            else:
                yaml.dump(configuration, conffile, encoding='utf-8')

            try:
                yield
            finally:
                self.forklift_class.configuration_file_list.pop()

    @staticmethod
    def localhost_reference():
        """
        The local host, as seen from inside the driver.
        """
        return 'localhost'

    def test_basic_environment(self):
        """
        Test passing basic environment to the command.
        """

        env = self.capture_env()
        self.assertEqual(env['DEVNAME'], 'myself')
        self.assertEqual(env['ENVIRONMENT'], 'dev_local')
        self.assertEqual(env['SITE_PROTOCOL'], 'http')
        self.assertTrue(re.match(r'^localhost:\d+$', env['SITE_DOMAIN']))

        env = self.capture_env('--serve_port', '9998')
        self.assertEqual(env['SITE_DOMAIN'], 'localhost:9998')

    def test_service_environment(self):
        """
        Test passing service environment to the command.
        """

        with self.configuration_file({'services': ['test']}):
            self.assertEqual(
                self.capture_env()['FOO'],
                '{0}-test_app-2'.format(self.localhost_reference())
            )

            empty_file = \
                """
                # An empty YAML file.
                """
            with self.configuration_file(empty_file):
                self.assertEqual(
                    self.capture_env()['FOO'],
                    '{0}-test_app-2'.format(self.localhost_reference())
                )

        with self.configuration_file({
            'services': ['test'],
            'test': {
                'one': '111',
            },
        }):
            self.assertEqual(self.capture_env()['FOO'],
                             '{0}-111-2'.format(self.localhost_reference()))

            with self.configuration_file({
                'test': {
                    'two': '222',
                },
            }):
                self.assertEqual(
                    self.capture_env()['FOO'],
                    '{0}-111-222'.format(self.localhost_reference()))

                self.assertEqual(
                    self.capture_env('--test.host', 'otherhost')['FOO'],
                    'otherhost-111-222'
                )

    def test_nargs(self):
        """
        Test multiple arguments
        """

        with self.configuration_file({
            'services': ['test'],
            'test': {
                'list': ['1', '2'],
            },
        }):
            self.assertEqual(
                self.capture_env()['BAR'],
                '1|2')

    def test_added_environment(self):
        """
        Test passing additional environment to the command.
        """

        with self.configuration_file({
            'environment': [
                'BAR=additional',
            ],
        }):
            self.assertEqual(self.capture_env()['BAR'], 'additional')

        # Environment can be passed as a hash
        with self.configuration_file({
            'environment': {
                'BAR': 'additional',
            },
        }):
            self.assertEqual(self.capture_env()['BAR'], 'additional')


class DirectEnvironmentTestCase(CaptureEnvironmentMixin, TestCase):
    """
    Test that environment is passed to the commands using direct driver.
    """

    default_driver = 'direct'


@requires_docker
class DockerEnvironmentTestCase(CaptureEnvironmentMixin, TestCase):
    """
    Test environment passed to the commands using Docker.
    """

    default_driver = 'docker'

    @staticmethod
    def localhost_reference():
        # TODO: Can this change?
        return ip_address('docker0')

    def capture_env(self, *args, prepend_args=None):
        """
        Run Forklift to capture the environment.
        """

        prepend_args = prepend_args or []
        prepend_args.append(DOCKER_BASE_IMAGE)

        args = ['--rm'] + list(args)

        return super().capture_env(*args, prepend_args=prepend_args)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Services that can be provided to running applications - base definitions.
"""

import logging
import os
import socket
import subprocess
import sys
import urllib.parse

from collections import namedtuple
from operator import attrgetter

import docker

import requests.exceptions

from xdg.BaseDirectory import save_cache_path

from forklift.base import ImproperlyConfigured, wait_for, rm_tree_root_owned
from forklift.registry import Registry

LOGGER = logging.getLogger(__name__)
register = Registry()  # pylint:disable=invalid-name


def try_port(host, port):
    """
    Try to connect to a given TCP port.
    """

    with socket.socket() as sock:
        sock.connect((host, int(port)))
        return True


def port_open(host, port):
    """
    Check whether the specified TCP port is open.
    """

    try:
        return try_port(host, port)
    except socket.error:
        return False


def split_host_port(host_port, default_port):
    """
    Split host:port into host and port, using the default port in case
    it's not given.
    """

    host_port = host_port.split(':')
    if len(host_port) == 2:
        host, port = host_port
        return host, port
    else:
        return host_port[0], default_port


def pipe_split(value):
    """
    Split a pipe-separated string if it's the only value in an array.
    """

    if isinstance(value, str):
        value = (value,)

    try:
        value = tuple(value)
    except TypeError:
        value = (value,)

    if len(value) == 1 and isinstance(value[0], str) and '|' in value[0]:
        return value[0].split('|')
    else:
        return value


def transient_provider(func):
    """
    Decorator to mark a provider as transient
    """
    func.transient = True
    return func


class Service(object):
    """
    Base class for services required by the application.
    """

    # A list of class methods to try to find an available service provider.
    providers = ()

    # A list of attributes which can be overridden from a configuration file
    # or the command line.
    allow_override = ()

    # A list of attributes which can be overridden as a list of arguments
    # (i.e. hosts, urls)
    allow_override_list = ()

    TEMPORARY_AVAILABILITY_ERRORS = ()
    PERMANENT_AVAILABILITY_ERRORS = ()

    # invalid-name disabled to allow it to conform with other availability
    # areas constants
    @property
    def AVAILABILITY_ERRORS(self):  # pylint:disable=invalid-name
        """
        Combine all availability errors
        """
        return (self.TEMPORARY_AVAILABILITY_ERRORS +
                self.PERMANENT_AVAILABILITY_ERRORS)

    CONTAINER_IMAGE = None
    DEFAULT_PORT = None

    @classmethod
    def add_arguments(cls, add_argument):
        """
        Add service configuration arguments to the parser.
        """

        # TODO: refactor for types other than string (port numbers) and
        # list (Elasticsearch host).

        for param in cls.allow_override:
            add_argument('--{0}'.format(param))

        for param in cls.allow_override_list:
            add_argument('--{0}'.format(param), nargs='+')

    @classmethod
    def provide(cls, application_id, overrides=None, transient=False):
        """
        Choose the first available service from the list of providers.
        """
        for provider in cls.providers:
            provider_func = getattr(cls, provider)
            if transient and not getattr(provider_func, 'transient', False):
                LOGGER.debug("Skipping %s provider for %s service because "
                             "it's not transient", provider, cls.__name__)
                continue

            LOGGER.debug("Trying %s provider for %s service",
                         provider, cls.__name__)
            try:
                service = provider_func(application_id)
                setattr(service, 'provided_by', provider)
            except ProviderNotAvailable as exc:
                print((
                    "While trying '{provider}' provider for {service}: {exc}"
                ).format(
                    provider=provider,
                    service=cls.__name__,
                    exc=exc,
                ), file=sys.stderr)
                continue

            cls._set_overrides(service, overrides)

            try:
                if service.available():
                    return service
            except:
                service.cleanup()
                raise

        raise ImproperlyConfigured(
            "No available providers for service {0}.".format(cls.__name__))

    @classmethod
    def _set_overrides(cls, service, overrides=None):
        """
        Setup override values on a service
        """
        overrides = vars(overrides) if overrides else {}
        allowed_overrides = cls.allow_override + cls.allow_override_list

        for key, value in overrides.items():
            if value is not None:
                if key in allowed_overrides:
                    setattr(service, key, value)
                    LOGGER.debug("Config for %s: %s = %s",
                                 cls.__name__, key, value)
                else:
                    raise ImproperlyConfigured(
                        "Invalid parameter {0} for service {1}.".format(
                            key, cls.__name__))

    def available(self):
        """
        Wrap check_available so that "expected" exceptions are not raised
        """
        try:
            return self.check_available()
        except self.AVAILABILITY_ERRORS:
            return False

    def check_available(self):
        """
        Check whether the service is available. Override to implement
        availability checks to warn the user instead of let the application
        fail.
        """
        return True

    def wait_until_available(self, retries=60):
        """
        Wait for the container to be available before returning. If the retry
        limit is exceeded, ProviderNotAvailable is raised

        Parameters:
            retries - number of times to retry before giving up
        """
        try:
            LOGGER.info("Waiting for %s to become available",
                        self.__class__.__name__)
            available = wait_for(
                self.check_available,
                expected_exceptions=self.TEMPORARY_AVAILABILITY_ERRORS,
                retries=retries,
            )
            return available

        except self.PERMANENT_AVAILABILITY_ERRORS as ex:
            print("Error checking for {}: {}".format(
                self.__class__.__name__, ex
            ))
            return False

    def environment(self):
        """
        The environment, as a dictionary, to let the application know
        the service configuration.
        """

        raise NotImplementedError("Please override environment().")

    def cleanup(self):
        """
        Do any clean up required to undo anything that was done in the provide
        method
        """

        # pylint:disable=no-member
        if self.provided_by != 'container':
            LOGGER.debug("Don't know how to clean up %s service provided "
                         "by %s",
                         self.__class__.__name__,
                         self.provided_by)
            return False

        if self.container_info.new:
            LOGGER.debug("Cleaning up container '%s' for %s service",
                         self.container_info.name,
                         self.__class__.__name__)
            destroy_container(self.container_info.name)
        else:
            LOGGER.debug("Not cleaning up container '%s' for service %s "
                         "because it was not created by this invocation",
                         self.container_info.name,
                         self.__class__.__name__)

        return True

    @classmethod
    def ensure_container(cls, application_id, **kwargs):
        """
        Ensure a container for this service is running.
        """

        return _ensure_container(
            image=cls.CONTAINER_IMAGE,
            port=cls.DEFAULT_PORT,
            application_id=application_id,
            **kwargs
        )

    @classmethod
    def from_container(cls, application_id, container):
        """
        The service instance connecting to the specified Docker container.
        """

        raise NotImplementedError("Please override from_container.")

    @classmethod
    @transient_provider
    def container(cls, application_id):
        """
        A generic container provider for a service. Needs to be specified in
        'providers' to activate.
        """

        container = cls.ensure_container(application_id)

        instance = cls.from_container(application_id, container)

        # pylint:disable=attribute-defined-outside-init
        instance.container_info = container
        return instance


def replace_part(url, **kwargs):
    """
    Replace a part of the URL with a new value.

    Keyword arguments can be any properties of urllib.parse.ParseResult.
    """

    netloc_parts = ('username', 'password', 'hostname', 'port')

    for part, value in kwargs.items():
        if part in netloc_parts:
            # Reassemble netloc
            netloc = {
                p: getattr(url, p)
                for p in netloc_parts
            }

            netloc[part] = value

            netloc_str = netloc['hostname']

            if netloc['port']:
                netloc_str += ':' + str(netloc['port'])
            if netloc['username'] or netloc['password']:
                userinfo = netloc['username'] or ''
                if netloc['password']:
                    userinfo += ':' + netloc['password']
                netloc_str = userinfo + '@' + netloc_str

            kwargs = {'netloc': netloc_str}
        else:
            kwargs = {part: value}

        # pylint:disable=protected-access
        url = url._replace(**kwargs)

    return url


class URLDescriptor(object):
    """
    A descriptor to get or set an URL part for all the URLs of the class.
    """

    def __init__(self, part, default='', joiner='|'.join):
        """
        Initialise a descriptor to get or set an URL part.

        Parameters:
            part - the part to get, can be a string or a tuple of (get, set)
            default - filler for missing parts of the URL, defaults to ''
            joiner - how to join the parts from the URL array together;
            defaults to concatenating with '|' in between
        """

        if isinstance(part, str):
            self.getter = attrgetter(part)
            self.setter = lambda url, value: replace_part(url, **{part: value})
        else:
            (self.getter, self.setter) = part

        self.default = default
        self.joiner = joiner

    def __get__(self, instance, owner):
        """
        Get the URL part for all URLs in the instance.
        """

        if instance is None:
            return self

        return self.joiner(
            self.getter(url) or self.default
            for url in instance.urls
        )

    def __set__(self, instance, value):
        """
        Set the URL part for all URLs in the instance.
        """

        instance.urls = tuple(
            self.setter(url, value)
            for url in instance.urls
        )


class URLNameDescriptor(URLDescriptor):
    """
    A descriptor to get or set the URL path without a leading slash,
    commonly used when mapping an alphanumeric namespace (e.g. database names)
    onto URLs.
    """

    def __init__(self):
        super().__init__((
            lambda url: url.path.lstrip('/'),
            lambda url, value: replace_part(url, path='/' + value),
        ))


class URLMultiValueDescriptor(URLDescriptor):
    """
    A descriptor to get or set part of the URLs to an array of values.
    """

    def __init__(self, part, default='', joiner=lambda xs: next(iter(xs))):
        super().__init__(part, default, joiner)

    def __set__(self, instance, value):
        """
        Set the URL part to an array of values.

        After setting this, all the URLs will be identical except for hostinfo
        taken from the (iterable) value assigned. The rest of the URL
        parameters will be copied from the first one.
        """

        url = instance.urls[0]
        instance.urls = tuple(
            self.setter(url, v)
            for v in pipe_split(value)
        )


class URLHostInfoDescriptor(URLMultiValueDescriptor):
    """
    A descriptor to get or set hostinfo pairs (hostname:port), accounting for
    default port.
    """

    def get_hostinfo(self, url):
        """
        Get the hostname:port pair of the URL.
        """

        port = url.port
        if port == self.default_port or port is None:
            return url.hostname
        else:
            return ':'.join((url.hostname, str(port)))

    def set_hostinfo(self, url, value):
        """
        Set the hostname:port pair of the URL.
        """

        hostname, port = split_host_port(value, self.default_port)
        return replace_part(url, hostname=hostname, port=port)

    def __init__(self, default_port, joiner=lambda xs: next(iter(xs))):
        self.default_port = default_port
        super().__init__(
            (
                self.get_hostinfo,
                self.set_hostinfo,
            ),
            joiner=joiner,
        )


class URLService(Service):
    """
    A service specified by a set of URLs.

    This is a common 12 factor pattern and should be used instead of inheriting
    Service directly as much as possible.
    """

    # These set respective attributes on all the URLs.
    allow_override = ('user', 'password', 'host', 'port', 'path')

    allow_override_list = ('urls',)

    def __init__(self, urls):
        self._urls = ()
        self.urls = urls

        log_service_settings(LOGGER, self, 'urls')

    def url_string(self):
        """
        All URLs joined as a string.
        """
        return '|'.join(url.geturl() for url in self.urls)

    @property
    def urls(self):
        """
        The array of the URLs the service can be accessed at.
        """

        return self._urls

    @urls.setter
    def urls(self, urls):
        """
        Set the URLs to access the service at.
        """

        self._urls = tuple(
            urllib.parse.urlparse(url) if isinstance(url, str) else url
            for url in pipe_split(urls)
        )

    user = URLDescriptor('username')
    password = URLDescriptor('password')
    host = hostname = URLMultiValueDescriptor('hostname')
    port = URLMultiValueDescriptor('port', default=None)
    path = URLDescriptor('path')


class ProviderNotAvailable(Exception):
    """
    A service provider is not available.
    """

    pass


class DependencyRequired(ProviderNotAvailable):
    """
    A dependency is required to make a provider available.
    """

    def __init__(self, message, command=None):
        super().__init__(message)
        self.command = command


class DockerImageRequired(DependencyRequired):
    """
    A Docker image is required to make a provider available.
    """

    def __init__(self, image):
        super().__init__(
            message="Docker image {0} is required.".format(image),
            command='docker pull {0}'.format(image),
        )


class ContainerRefusingConnection(ProviderNotAvailable):
    """
    A Docker container that was started is not connectable after a period of
    time.
    """

    def __init__(self, image, port):
        super().__init__(
            message=("Docker container {0} was started but couldn't connect on"
                     "port {1}").format(image, port)
        )


ContainerInfo = namedtuple('ContainerInfo', ['host',
                                             'port',
                                             'data_dir',
                                             'name',
                                             'new'])


def cache_directory(container_name):
    """
    A directory to cache the container data in.
    """

    return os.path.join(save_cache_path('forklift'), container_name)


def container_name_for(image, application_id):
    """
    Get a name for a service container based on image and application ID

    Parameters:
        image - image that the container is for
        application_id - application id that the container is for

    Return value:
        A string
    """
    return image.replace('/', '_') + '__' + application_id


def _ensure_container(image,
                      port,
                      application_id,
                      data_dir=None,
                      **kwargs):
    """
    Ensure that a container for an application is running and wait for the port
    to be connectable.

    Parameters:
        image - the image to run a container from
        port - the port to forward from the container
        application_id - the application ID, for naming the container
        data_dir - the directory to persistently mount inside the container

    Return value:
        An object with the following attributes:
            port - the forwarded port number
            data_dir - if asked for, path for the persistently mounted
            directory inside the container
            name - the container name
            new - True/False to show if the container was created or not
    """

    with docker.Client() as docker_client:

        # TODO: better container name
        container_name = container_name_for(image, application_id)
        LOGGER.info("Ensuring container for '%s' is started with name '%s'",
                    image, container_name)

        if data_dir is not None:
            cached_dir = cache_directory(container_name)
        else:
            cached_dir = None

        try:
            created, container_status = get_or_create_container(
                docker_client,
                container_name,
                image,
                port,
                data_dir,
                cached_dir,
                **kwargs
            )

            if not container_status['State']['Running']:
                _start_container(docker_client,
                                 container_name,
                                 port,
                                 data_dir,
                                 cached_dir)

            host = docker_client.inspect_container(
                container_name)['NetworkSettings']['IPAddress']
            host_port = docker_client.port(container_name, port)[0]['HostPort']

            try:
                _wait_for_port(image, host_port)
            except:
                if created:
                    LOGGER.debug("Could not connect to '%s' container, so "
                                 "destroying it", image)
                    destroy_container(container_name)
                raise

            return ContainerInfo(host=host,
                                 port=port,
                                 data_dir=cached_dir,
                                 name=container_name,
                                 new=created)
        except requests.exceptions.ConnectionError:
            raise ProviderNotAvailable("Cannot connect to Docker daemon.")


# pylint:disable=too-many-arguments
def get_or_create_container(docker_client,
                            container_name,
                            image,
                            port,
                            data_dir=None,
                            cached_dir=None,
                            **kwargs):
    """
    Get info for an existing container by name, or create a new one

    Parameters:
        docker_client - a docker.Client object for the Docker daemon
        container_name - name to check/start
        image - the image to run a container from
        port - the port to forward from the container
        data_dir - the directory to persistently mount inside the container
        cached_dir - the directory to mount from the host to data_dir

    Return value:
        A tuple of:
            - True if the container started as a result of this call
            - Output from Docker inspect
    """
    try:
        return False, docker_client.inspect_container(container_name)
    except docker.errors.APIError:
        try:
            docker_client.inspect_image(image)
        except docker.errors.APIError:
            raise DockerImageRequired(image)

        if data_dir is not None:
            # Ensure the data volume is mounted
            kwargs.setdefault('volumes', {})[data_dir] = cached_dir

        docker_client.create_container(
            image,
            name=container_name,
            ports=(port,),
            **kwargs
        )
        container_status = docker_client.inspect_container(container_name)

    return True, container_status


def destroy_container(container_name):
    """
    Stop and remove a container by name
    """
    cache_dir = cache_directory(container_name)
    with docker.Client() as docker_client:
        docker_client.stop(container_name)
        docker_client.remove_container(container_name)

    try:
        rm_tree_root_owned(cache_dir)
    except subprocess.CalledProcessError:
        pass


def _wait_for_port(image, port, retries=30):
    """
    Wait for a port to become available, or raise ContainerRefusingConnection
    error

    Parameters:
        image - the image that the container is run from
        port - the port to wait for
        retries - number of times to retry before giving up
    """
    LOGGER.debug("Waiting for '%s' port %s to be reachable", image, port)
    if not wait_for(lambda: port_open('127.0.0.1', port), retries=retries):
        raise ContainerRefusingConnection(image, port)


def _start_container(docker_client, image, port, data_dir, cached_dir):
    """
    Start a container, binding ports and data dirs

    Parameters:
        docker_client - client for the Docker API
        image - the image to run a container from
        port - the port to forward from the container
        data_dir - the directory to persistently mount inside the container
        cached_dir - the directory to mount from the host to data_dir
    """
    LOGGER.info("Starting '%s' container", image)
    LOGGER.debug("Container port: %s", port)
    LOGGER.debug("Container data dir (in container): %s", data_dir)
    LOGGER.debug("Container cached dir (on host): %s", cached_dir)
    start_args = {
        'port_bindings': {port: None},
    }
    if data_dir is not None:
        start_args['binds'] = {
            cached_dir: data_dir,
        }
    docker_client.start(image, **start_args)


def log_service_settings(logger, service, *attrs):
    """
    Format and log a service settings.

    Parameters:
        logger - a logger object to log to
        service - the service object that the settings are for
        attrs - a list of attrs to get from the service. If the attr is
                callable, it will be called with no arguments. It may return
                just a value, or a tuple of a new attr name and a value
    """
    if logger.isEnabledFor(logging.DEBUG):
        for attr in attrs:
            val = getattr(service, attr)
            if callable(val):
                val = val()

            logger.debug("%s %s: %s", service.__class__.__name__, attr, val)



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the --rm and --transient flags
"""

import docker
import docker.errors

from tests.base import (
    requires_docker_image,
    TestCase,
    TestDriver,
    TestForklift,
)


def assertion_driver(func):
    """
    Create a test driver that runs func to assert test conditions
    """
    class InnerClass(TestDriver):  # pylint:disable=missing-docstring
        run = func

    return InnerClass


def assertion_forklift_class(func):
    """
    Create a test forklift class that has only an assertion driver
    """
    class InnerClass(TestForklift):  # pylint:disable=missing-docstring
        drivers = {'assertion_driver': assertion_driver(func)}

    return InnerClass


@requires_docker_image('thatpanda/postgis')
class TestRm(TestCase):
    """
    Test the --rm flag
    """

    def setUp(self):
        self.client = docker.Client()

    def tearDown(self):
        self.client.close()

    def test_create_delete(self):
        """
        Make sure the container and data dirs are both created and destroyed
        """
        test_info = {}

        def assertions_func(driver, *_):
            """
            Test container/data dir creation
            """
            container_info = driver.services[0].container_info
            test_info['container_name'] = container_info.name
            test_info['data_dir'] = container_info.data_dir

            # FIXME: data_dir is broken in paintedfox container
            # self.assertTrue(os.path.isdir(test_info['data_dir']),
            #                 "Data dir is created")

            container_inspect = self.client.inspect_container(
                test_info['container_name'])
            self.assertTrue(container_inspect['State']['Running'],
                            "Container is running")

            return 0

        self.forklift_class = assertion_forklift_class(assertions_func)
        self.assertEqual(0, self.run_forklift(
            '--driver', 'assertion_driver',
            '--service', 'postgis',
            '--rm',
            '--', 'fake',
        ))

        # FIXME: data_dir is broken in paintedfox container
        # self.assertFalse(os.path.exists(test_info['data_dir']),
        #                  "Data dir does not exist")

        with self.assertRaises(docker.errors.APIError) as ex:
            self.client.inspect_container(test_info['container_name'])

        ex = ex.exception
        self.assertEqual(ex.response.status_code, 404)
        self.assertRegex(ex.explanation, b'No such container')


class TestTransient(TestCase):
    """
    Test the --transient flag
    """

    def test_transient_service(self):
        """
        Make sure that a transient service is selected
        """
        def assertions_func(driver, *_):
            """
            Make sure that the transient service is used
            """
            self.assertEqual('here_occasionally',
                             driver.services[0].provided_by)
            return 0

        self.forklift_class = assertion_forklift_class(assertions_func)
        self.assertEqual(0, self.run_forklift(
            '--driver', 'assertion_driver',
            '--service', 'test',
            '--transient',
            '--', 'fake',
        ))

    def test_non_transient_service(self):
        """
        Negative test to make sure the test_transient_service test is valid
        """
        def assertions_func(driver, *_):
            """
            Make sure that the transient service is not used
            """
            self.assertEqual('here',
                             driver.services[0].provided_by)
            return 0

        self.forklift_class = assertion_forklift_class(assertions_func)
        self.assertEqual(0, self.run_forklift(
            '--driver', 'assertion_driver',
            '--service', 'test',
            '--', 'fake',
        ))



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Setup script.
"""

from setuptools import setup, find_packages
from sys import version_info

assert version_info >= (3,), "Python 3 is required."

with open('requirements.txt') as requirements, \
        open('test_requirements.txt') as test_requirements:
    setup(
        name='docker-forklift',
        version='0.2.34',
        description='Utility for running a container',
        author='Infoxchange Australia development team',
        author_email='devs@infoxchange.net.au',
        url='https://github.com/infoxchange/docker-forklift',
        license='Apache 2.0',
        long_description=open('README.md').read(),

        packages=find_packages(exclude=['tests']),
        package_data={
            'forklift': [
                'README.md',
                'requirements.txt',
                'test_requirements.txt',
            ],
        },
        entry_points={
            'console_scripts': [
                'forklift = forklift:main',
            ],
        },

        install_requires=requirements.read().splitlines(),

        test_suite='tests',
        tests_require=test_requirements.read().splitlines(),
    )



#
# Copyright 2014  Infoxchange Australia
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Proxy service.
"""

import socket

from forklift.base import free_port
from .base import register, transient_provider, try_port, URLService


@register('email')
class Email(URLService):
    """
    An MTA for the application.
    """

    TEMPORARY_AVAILABILITY_ERRORS = (socket.error,)

    allow_override = ('host', 'port')

    def __init__(self, host, port=25):
        super().__init__((
            'smtp://{host}:{port}/'.format(
                host=host,
                port=port,
            ),
        ))

    def environment(self):
        """
        The environment to send email.
        """

        return {
            'EMAIL_HOST': self.host,
            'EMAIL_PORT': str(self.port),
        }

    def check_available(self):
        """
        Check whether the MTA is available.
        """

        return try_port(self.host, self.port)

    # pylint:disable=unused-argument
    @classmethod
    def localhost(cls, application_id):
        """
        The MTA on the local machine.
        """
        return cls(host='localhost')

    # pylint:disable=unused-argument
    @classmethod
    @transient_provider
    def stdout(cls, application_id):
        """
        Mailer printing all the messages to the standard output of Forklift.
        """

        from forklift.services.satellite import start_satellite
        from smtpd import DebuggingServer

        port = free_port()

        def run_server():
            """
            Run the syslog server.
            """

            DebuggingServer(('0.0.0.0', port), None)
            import asyncore
            asyncore.loop()

        start_satellite(target=run_server)

        instance = cls('localhost', port)
        instance.wait_until_available()
        return instance

    providers = ('localhost', 'stdout')
